/*
 * fmradio.c - FM Radio for Commodore 64
 *
 * Copyright (c) 2013 Timo Voutilainen
 *
 * This file is part of FM Radio for Commodore 64.
 *
 * FM Radio for Commodore 64 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FM Radio for Commodore 64 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FM Radio for Commodore 64.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Thanks to Maarten ter Huurne for a reference Linux Kernel driver.
 */


#define __CBM__ 1

//#define DEBUG 1
//#define VICE 1
//#define ADDFEATURES 1

#define VERSION	"0.3"

#include <stdio.h>
#include <peekpoke.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <c64.h>
#include <ctype.h>
#include "i2c.h"

#define RADIOADDR	0x11
#define NO 0x00
#define MONO 0
#define STEREO 1
#define SET_STEREO 0
#define SET_MONO 1
#define CONFIG_NOT_CHANGED 	0
#define CONFIG_CHANGED 		1
#define CFG_CHA_MAGIC		"chan2st"

enum rda5807_register_addresses {
	RDA5807_CHIPID_REG			= 0x00,
	RDA5807_CTRL_REG			= 0x02,
	RDA5807_CHANNEL_REG			= 0x03,
	RDA5807_IOCFG_REG			= 0x04,
	RDA5807_INTM_THRESH_VOL_REG	= 0x05,
	RDA5807_SEEK_RESULT_REG		= 0x0A,
	RDA5807_SIGNAL_REG			= 0x0B,
};

#define TIMER_EXPIRED 0x00
#define TIMER_RUNNING 0x01
#define TIMERBHIGH 0xDC07
#define TIMERBLOW  0xDC06
#define TIMERBCONTROL 0xDC0F
#define TIMERSTARTBIT_MASK 0x01

#define RDA5807_MASK_CTRL_HIBYTE_DHIZ		0x80 //BIT(7) //BIT(15)
#define RDA5807_MASK_CTRL_HIBYTE_DMUTE		0x40 //BIT(6) //BIT(14)
#define RDA5807_MASK_CTRL_HIBYTE_MONO		0x20 //BIT(5) //BIT(13)
#define RDA5807_MASK_CTRL_HIBYTE_BASS		0x10 //BIT(4) //BIT(12)
#define RDA5807_MASK_CTRL_HIBYTE_SEEKUP 	0x02 //BIT(1) //BIT(9)
#define RDA5807_MASK_CTRL_HIBYTE_SEEK		0x01 //BIT(0) //BIT(8)
#define RDA5807_MASK_CTRL_LOBYTE_SKMODE		0x80 //BIT(7)
#define RDA5807_MASK_CTRL_LOBYTE_CLKMODE	(7 << 4)
#define RDA5807_MASK_CTRL_LOBYTE_SOFTRESET	0x02 //BIT(1)
#define RDA5807_MASK_CTRL_LOBYTE_ENABLE		0x01 //BIT(0)

#define RDA5807_SHIFT_CHAN_WRCHAN    		6
#define RDA5807_MASK_CHAN_WRCHAN    		(0x3FF << RDA5807_SHIFT_CHAN_WRCHAN)
#define RDA5807_MASK_CHAN_TUNE        		0x10 //BIT(4)
#define RDA5807_SHIFT_CHAN_BAND        		2
#define RDA5807_MASK_CHAN_BAND        		(0x3 << RDA5807_SHIFT_CHAN_BAND)
#define RDA5807_SHIFT_CHAN_SPACE    		0
#define RDA5807_MASK_CHAN_SPACE        		(0x3 << RDA5807_SHIFT_CHAN_SPACE)

#define RDA5807_MASK_SEEKRES_HIBYTE_COMPLETE   	0x40 //BIT(6) //BIT(14)
#define RDA5807_MASK_SEEKRES_HIBYTE_FAIL   		0x20 //BIT(5) //BIT(13)
#define RDA5807_MASK_SEEKRES_HIBYTE_STEREO  	0x04 //BIT(2) //BIT(10)
#define RDA5807_MASK_HIBYTE_DEEMPHASIS     		0x08 //BIT(3) //BIT(11)
#define RDA5807_MASK_LOBYTE_VOLUME_DAC  	0xF
#define RDA5807_MASK_LOBYTE_LNA_PORT		0xC0
#define RDA5807_MASK_LOBYTE_LNA_CURRENT		0x03

#define RDA5807_SHIFT_RSSI        			9
#define RDA5807_MASK_RSSI        			(0x7F << RDA5807_SHIFT_RSSI)

#define RDA5807_FREQ_MIN_100KHZ				870
#define RDA5807_FREQ_MAX_100KHZ				1080

#define RDA5807_FREQ_TEST_FREQ				9040

#define RDA5807_CHIP_ID						0x58

#define RDA5807_VOLUME_MIN					0x00
#define RDA5807_VOLUME_DEFAULT				0x0A
#define RDA5807_VOLUME_MAX					0x0F

#define RDA5807_DEEMPHASIS_50_US			0x08
#define RDA5807_DEEMPHASIS_75_US			0x00

#define RDA5807_LNA_PORT_NOINPUT			0x00
#define RDA5807_LNA_PORT_LNAN				0x40
#define RDA5807_LNA_PORT_LNAP				0x80
#define RDA5807_LNA_PORT_DUAL				0xC0
#define RDA5807_LNA_CURRENT_1_8MA			0x00
#define RDA5807_LNA_CURRENT_2_1MA			0x01
#define RDA5807_LNA_CURRENT_2_5MA			0x02
#define RDA5807_LNA_CURRENT3_0MA			0x03

#define MUTE_ON			1
#define MUTE_OFF		0
#define OFF				0
#define VOLUME_BAR_X	14
#define VOLUME_BAR_Y	3
#define RSSI_BAR_X		16
#define RSSI_BAR_Y		3
#define VOLUME_TEXT_X	14
#define VOLUME_TEXT_Y	5
#define RSSI_TEXT_X		16
#define RSSI_TEXT_Y		5
#define CHAN_INFO_X		22
#define SEEK_DOWN		0
#define SEEK_UP			1
#define MAX_CHAN_AMOUNT 10
#define MAX_CHAN_NAME_LEN 15

enum radio_states {
	RADIO_OFF = 0,
	RADIO_ON,
	RADIO_SCANNING,
	RADIO_ASK_CHAN_NO,
	RADIO_ASK_CHAN_NAME,
	RADIO_QUIT_ASK_SAVE
};

typedef enum exit_save_status {
	SAVED_OK= 0,
	SAVE_FAILED,
	SAVE_NOT_REQUESTED,
}exit_save_status;

typedef struct chan_str
{
	u8     chan_number;
	u16    chan_freq;
	u8     name[MAX_CHAN_NAME_LEN];
};

static struct chan_str channels[MAX_CHAN_AMOUNT];


static u8 radio_state = RADIO_OFF;
static u8 volume = RDA5807_VOLUME_DEFAULT;
static u8 bass_boost = OFF;
static u16 radio_freq = RDA5807_FREQ_MIN_100KHZ;
static u8 rssi = 0;
static u8 a_char_to_display; /* Used by ASM in put_petscii() ... needs a global variable*/
static u8 fm_is_channel = OFF;
static u8 fm_seek_ready = OFF;
static u8 timer_ticks = 0;
static u8 radio_mute = OFF;
static u8 seek_or_tune_complete = OFF;
static u8 seek_fail = OK;
static u8 stereo_indicator = OFF;
static u8 set_stereo = SET_STEREO;
static u8 conf_changed = CONFIG_NOT_CHANGED;
static u8 deemphasis = RDA5807_DEEMPHASIS_50_US;
#ifdef ADDFEATURES

static u8 lna_port_states[4] = {RDA5807_LNA_PORT_NOINPUT,
					RDA5807_LNA_PORT_LNAN,
					RDA5807_LNA_PORT_LNAP,
					RDA5807_LNA_PORT_DUAL};
static u8 lna_port_state = 1;

static u8 lna_current_states[4] = {RDA5807_LNA_CURRENT_1_8MA,
						RDA5807_LNA_CURRENT_2_1MA,
						RDA5807_LNA_CURRENT_2_5MA,
						RDA5807_LNA_CURRENT3_0MA};
static u8 lna_current_state = 1;
#endif


/* Function Prototypes */

static u8 init_fmradio();
static u8 set_mute(u8 state);
static u8 set_frequency(u16 *freq_100khz);
static u8 set_radio_state(u8 state);
static u8 set_volume(u8 vol);
static void handle_keypressed();
static void draw_layout_skeleton();
static u8 toggle_bass_boost();
static void put_petscii(u8 a_char);
static void update_power_status();
static void update_freq_display();
static void update_vert_bar(u8 value, u8 x, u8 y);
static void update_bass_boost_view();
static void update_mute_view();
static void update_scan_view();
static void update_channel_status_view();
static void update_channel_name_view();
static void update_channel_quick_sel_view();
static void update_set_stereo_status();
static u8 get_rssi_and_stat();
static u16 get_frequency();
static u8 get_stc_sf_and_st();
static u8 set_stereo_state(u8 set_stereo);
static void start_timer();
static u8 is_timer_complete(u8 ticks);
static void print_registers();
static u8 start_scan(u8 direction);
static u8 save_config_file();
static u8 read_config_file();
static void tune_to_first_stored_slot();
static void exit_program(exit_save_status exit_status);
static u8 set_deemphasis(u8 deemphasis);
static void update_deemphasis_status();

#ifdef ADDFEATURES
static u8 set_lna_port(u8 lna_port);
static u8 set_lna_current(u8 lna_current);
#endif

u8 main(void)
{
    static u16 prev_freq = 0;
    u8 radio_init_retries = 0;

    if (init_i2c_bus() != OK)
    {
    	printf("I2C Init failed!\n");
    	exit(1);
    }
    while (radio_init_retries < 5)
    {
    	if (init_fmradio() == OK)
    		goto init_ok;
    	radio_init_retries++;
    }
	printf("Radio Init failed!\n");
	exit(1);

init_ok:
	clrscr();

#ifdef DEBUG
	gotoxy(1,19);
	printf("Init retries: %d", radio_init_retries);
#endif

    memset(channels, 0, sizeof(channels));
    bordercolor(0);
    bgcolor(0);
    put_petscii(8); // Disable shift + commodore -key
    put_petscii(14);/* Switch to lowercase */
    draw_layout_skeleton();
    update_power_status();
    update_freq_display();
    update_vert_bar(volume, VOLUME_BAR_X, VOLUME_BAR_Y);
    update_mute_view();
    update_bass_boost_view();
    update_scan_view();
    update_channel_status_view();
    update_set_stereo_status();
    update_channel_quick_sel_view();
    update_deemphasis_status();

    gotoxy(CHAN_INFO_X,7);
    textcolor(COLOR_YELLOW);
    printf("Reading config...");
    if (read_config_file() != OK)
    {
        memset(channels, 0, sizeof(channels));
        gotoxy(CHAN_INFO_X,7);
        printf("Read Conf Failed!");
    }
    else
    {
        gotoxy(CHAN_INFO_X,7);
    	printf("Config read OK   ");
    	tune_to_first_stored_slot();
    }
    update_freq_display();
    update_channel_quick_sel_view();
    start_timer();
    while (is_timer_complete(40) == TIMER_RUNNING);
    update_channel_name_view();

    start_timer();
    while (1)
    	{
    	if (radio_state == RADIO_ON || radio_state == RADIO_SCANNING)
    	{
        	if (is_timer_complete(15) == TIMER_EXPIRED)
        	{
        		get_stc_sf_and_st();
        		get_rssi_and_stat();
    			update_vert_bar(rssi/4, RSSI_BAR_X, RSSI_BAR_Y);
    			update_channel_status_view();
    			if (radio_state == RADIO_SCANNING && (fm_seek_ready || seek_fail == NOT_OK))
    			{
    				seek_fail = OK;
    				radio_freq = get_frequency();
    				update_freq_display();
    				radio_state = RADIO_ON;
    				update_scan_view();
    			}
    			if (radio_freq != prev_freq)
    			{
    			    update_channel_name_view();
    			    update_channel_quick_sel_view();
    			    prev_freq = radio_freq;
    			}
        	}
    	}

    	if (kbhit())
          {
          handle_keypressed();
          }
       }
    return OK;
}

/* --------------------------------------------------------------- */
static void handle_keypressed()
{
	u8 key;
	static signed char chan_no = 0;

	key = cgetc();

	// Radio ON and OFF -states
	if ((radio_state == RADIO_ON) ||
	    (radio_state == RADIO_OFF))
	{
		switch (key)
		{
				case 'Q':
				case 'q':
					if (conf_changed == CONFIG_CHANGED)
					{
						textcolor(COLOR_YELLOW);
						cclearxy(CHAN_INFO_X,6,17);
						cclearxy(CHAN_INFO_X,7,17);
						gotoxy(CHAN_INFO_X,6);
						printf("Save config Y/N ?");
						radio_state = RADIO_QUIT_ASK_SAVE;
					}
					else
						exit_program(SAVE_NOT_REQUESTED);

					break;
		}
	}


	if (radio_state == RADIO_ON)
		{
		if (isdigit(key))
		{
			if (channels[atoi(&key)].chan_freq != 0)
			{
				radio_freq = channels[atoi(&key)].chan_freq;
				set_frequency(&radio_freq);
				update_freq_display();
				return;
			}
		}
		switch (key)
			{
#ifdef ADDFEATURES
			case 'P':
			case 'p':
				lna_port_state++;
				if (lna_port_state >= sizeof(lna_port_states))
					lna_port_state = 0;
				set_lna_port(lna_port_states[lna_port_state]);
				break;
			case 'I':
			case 'i':
				lna_current_state++;
				if (lna_current_state >= sizeof(lna_current_states))
					lna_current_state = 0;
				set_lna_current(lna_current_states[lna_current_state]);
				break;
#endif
			case 'E':
			case 'e':
				if (deemphasis == RDA5807_DEEMPHASIS_75_US)
					deemphasis = RDA5807_DEEMPHASIS_50_US;
				else
					deemphasis = RDA5807_DEEMPHASIS_75_US;
				set_deemphasis(deemphasis);
				update_deemphasis_status();
				break;
			case 168:
				print_registers();
				break;

			case 'S':
			case 's':
				if (set_stereo == SET_STEREO)
					set_stereo = SET_MONO;
				else
					set_stereo = SET_STEREO;
				set_stereo_state(set_stereo);
				update_set_stereo_status();
				break;
			case 'C':
			case 'c':
				radio_state = RADIO_ASK_CHAN_NO;
				gotoxy(CHAN_INFO_X,6);
				textcolor(COLOR_GREEN);
				printf("Ch no(0-9)     ");
				cclearxy(CHAN_INFO_X,7,17);
				break;
			case ' ':
				radio_state = RADIO_OFF;
				set_radio_state(radio_state);
				update_power_status();
				break;
			case ':':
				radio_state = RADIO_SCANNING;
				update_scan_view();
				start_scan(SEEK_DOWN);
				break;
			case ';':
				radio_state = RADIO_SCANNING;
				update_scan_view();
				start_scan(SEEK_UP);
				break;
			case '+':
				volume++;
				set_volume(volume);
				update_vert_bar(volume, VOLUME_BAR_X, VOLUME_BAR_Y);
				break;
			case '-':
				volume--;
				set_volume(volume);
				update_vert_bar(volume, VOLUME_BAR_X, VOLUME_BAR_Y);
				break;
			case 'B':
			case 'b':
				toggle_bass_boost();
				update_bass_boost_view();
				break;
			case 'U':
			case 'u':
				radio_freq += 1;
				set_frequency(&radio_freq);
				update_freq_display();
				break;
			case 'D':
			case 'd':
				radio_freq -= 1;
				set_frequency(&radio_freq);
				update_freq_display();
				break;
			case 'M':
			case 'm':
				radio_mute ^= 1;
				set_mute(radio_mute);
				update_mute_view();
				break;
			}
		}
	else if (radio_state == RADIO_OFF)
		{
		switch (key)
			{
			case ' ':
				radio_state = RADIO_ON;
				set_radio_state(radio_state);
				update_power_status();
				break;
			}
		}
	else if (radio_state == RADIO_ASK_CHAN_NO)
		{
		gotoxy(CHAN_INFO_X,6);
		printf("         ");
		if (isdigit(key))
		{
			chan_no = atoi(&key);
			radio_state = RADIO_ASK_CHAN_NAME;
			cclearxy(CHAN_INFO_X,6,17);
			gotoxy(CHAN_INFO_X,6);
			printf("Ch name:");
			gotoxy(CHAN_INFO_X,7);
			memset(channels[chan_no].name, '\0', sizeof(channels[chan_no].name));
		}
		else //Canceled
		{
			radio_state = RADIO_ON;
		    cclearxy(CHAN_INFO_X,6,17);
			update_channel_name_view();
			update_channel_quick_sel_view();
		}
		}

	else if (radio_state == RADIO_ASK_CHAN_NAME)
		{
		if (isprint(key))
		{
			if (strlen(channels[chan_no].name) < (MAX_CHAN_NAME_LEN - 1))
			{
				channels[chan_no].name[strlen(channels[chan_no].name)] = key;
				putc(key, stdout);
			}
		}
		else if (key == 20) //Delete previous character
		{
			cclearxy(CHAN_INFO_X,7,17);
			//Move cursor to original start
			gotoxy(CHAN_INFO_X,7);
			//Delete last char from name
			if (strlen(channels[chan_no].name) > 0)
				channels[chan_no].name[strlen(channels[chan_no].name) - 1] = '\0';
			printf("%s", channels[chan_no].name);
		}
		else if (key == 13) //Accepted with enter
		{
			channels[chan_no].chan_freq = radio_freq;
			channels[chan_no].chan_number = chan_no;
			cclearxy(CHAN_INFO_X,6,17);
			update_channel_name_view();
			update_channel_quick_sel_view();
			radio_state = RADIO_ON;
			conf_changed = CONFIG_CHANGED;
		}

		}
	else if (radio_state == RADIO_QUIT_ASK_SAVE)
	{
		switch (key)
		{
		case 'y':
		case 'Y':
			if (save_config_file() == OK)
			{
				exit_program(SAVED_OK);
			}
			else
			{
				exit_program(SAVE_FAILED);
			}
			break;
		case 'n':
		case 'N':
			exit_program(SAVE_NOT_REQUESTED);
			break;
		}

	}
}

/* --------------------------------------------------------------- */
static void exit_program(exit_save_status exit_status)
{
    clrscr();
    put_petscii(142);/* Switch to Uppercase */
    if (exit_status == SAVED_OK)
    	printf("config saved ok.\n");
    else if (exit_status == SAVE_FAILED)
    	printf("config save failed!\n");
    printf("bye...");
	exit(0);
}
/* --------------------------------------------------------------- */
static void draw_layout_skeleton()
{
	u8 i;

    textcolor(COLOR_BLUE);
    revers(0);
    gotoxy(0,0);
    gotoxy(0,0);
    put_petscii(176);/* Upper left corner char */
    revers(1);

    printf("         C64 FM-radio (c) T.V. v%s    ", VERSION);
    revers(0);
    gotoxy(39,0);
    put_petscii(174);/* Upper right corner char */

    // Left and right walls
    for (i=1;i < 8;i++)
        {
        gotoxy(0,i);
        put_petscii(125);/* Vertical line in the middle */
        gotoxy(39,i);
        put_petscii(125);
        }

    for (i=0;i<38;i++)
        {
        gotoxy(i,8);
        put_petscii(96); /* Horizontal line in the middle */
        }
    gotoxy(38,8);
    put_petscii(189); /* Bottom right corner char */

    /* Move bottom line 1 char to the right with 'INS' */
    gotoxy(0,8);
    put_petscii(148);
    put_petscii(173); /* Bottom left corner char */

    /* Now the update the splitting vertical line */
    for (i=1;i<8;i++)
        {
        gotoxy(11,i);
        put_petscii(125);
        }
    gotoxy(11,8);
    put_petscii(177);

    gotoxy(VOLUME_TEXT_X, VOLUME_TEXT_Y);
    printf("V S");
    gotoxy(VOLUME_TEXT_X, VOLUME_TEXT_Y + 1);
    printf("O I");
    gotoxy(VOLUME_TEXT_X, VOLUME_TEXT_Y + 2);
    printf("L G");
}

/* --------------------------------------------------------------- */
static void update_power_status()
{
	gotoxy(1,1);
	revers(1);
	if (radio_state == RADIO_ON)
		{
		textcolor(COLOR_GREEN);
		}
	else
	{
		textcolor(COLOR_RED);
	}
	printf("  CPOWER  ");
	revers(0);
}

/* --------------------------------------------------------------- */
static void update_freq_display()
{
	gotoxy(22,2);
	textcolor(COLOR_RED);
	printf("%3d.%dMHz",radio_freq/10, radio_freq%10);
}

/* --------------------------------------------------------------- */
static void update_vert_bar(u8 value, u8 x, u8 y)
{
	u8 index = value;
	u8 chars[8][2] = {{32,146},{228,146},{239,146},{226,146},{248,18},{247,18},{227,18},{32,18}};

	gotoxy(x, y);
	textcolor(COLOR_GREEN);
	if (index > 7) //First char is full box
		index = 7;
	put_petscii(chars[index][1]);
	put_petscii(chars[index][0]);
	// Upper box...
	textcolor(COLOR_YELLOW);
	gotoxy(x, y - 1);
	if (value > 7)
		{
		index = value - 7;
		if (index > 7)
			index = 7;
		put_petscii(chars[index][1]);
		put_petscii(chars[index][0]);
		}
	else
		{ // No upper box
		put_petscii(chars[0][1]);
		put_petscii(chars[0][0]);
		}
	revers(0);
	gotoxy(x, y - 2);
	if (value > 14)
	{
		index = 1;
		textcolor(COLOR_RED);
		//put_petscii(chars[index][1]);
		put_petscii(chars[index][0]);
	}
	else
		put_petscii(32); //Empty space
}

/* --------------------------------------------------------------- */
static void update_mute_view()
{
	gotoxy(1, 3);
	textcolor(COLOR_GRAY1);
	revers(radio_mute);
	printf("   Mute   ");
	revers(0);
}

/* --------------------------------------------------------------- */
static void update_bass_boost_view()
{
	gotoxy(1, 4);
	textcolor(COLOR_GRAY2);
	revers(bass_boost);
	printf("Bass Boost");
	revers(0);
}

/* --------------------------------------------------------------- */
static void update_scan_view()
{
	gotoxy(1, 5);
	textcolor(COLOR_GRAY1);
	if (radio_state == RADIO_SCANNING)
		revers(1);
	printf("   Scan   ");
	revers(0);
}

/* --------------------------------------------------------------- */
static void update_set_stereo_status()
{
	gotoxy(1, 6);
	textcolor(COLOR_GRAY2);
	if (set_stereo == SET_STEREO)
	{
		revers(1);
		printf("  Stereo  ");
	}
	else
		printf("   Mono   ");

	revers(0);
}

/* --------------------------------------------------------------- */
static void update_deemphasis_status()
{
	gotoxy(1, 7);
	textcolor(COLOR_GRAY1);
	if (deemphasis == RDA5807_DEEMPHASIS_50_US)
	{
		printf("50us De-em");
	}
	else
		printf("75us De-em");
}

/* --------------------------------------------------------------- */
static void update_channel_status_view()
{
	gotoxy(18, 2);
	textcolor(COLOR_GRAY1-(fm_is_channel * 6)); //GREEN
	revers(fm_is_channel);
	printf("CH");

	gotoxy(18, 3);
	textcolor(COLOR_GRAY1-(stereo_indicator * 6)); //GREEN
	revers(stereo_indicator);
	printf("ST");
	revers(0);
}

/* --------------------------------------------------------------- */
static void update_channel_name_view()
{
	u8 i = 0, found = 0;

	textcolor(COLOR_BLUE);
	gotoxy(CHAN_INFO_X, 6);
	printf("Station name:");

	textcolor(COLOR_RED);
	cclearxy(CHAN_INFO_X, 7, 17);
	//Move cursor to original start
	gotoxy(CHAN_INFO_X, 7);

	for (i = 0; i < MAX_CHAN_AMOUNT; i++)
	{
		if (radio_freq == channels[i].chan_freq)
		{
			//printf("%d. %s", channels[i].chan_number, channels[i].name);
			printf("%s", channels[i].name);
			found = 1;
			break;
		}
	}
	if (!found)
		printf("--- ");

}
/* --------------------------------------------------------------- */
static void update_channel_quick_sel_view()
{
	u8 i = 1;

	gotoxy(CHAN_INFO_X, 4);
	//for (i = 1; ((i <= MAX_CHAN_AMOUNT) && (i != 0)); i++)
	while (i <= MAX_CHAN_AMOUNT)
	{
		//gotoxy(35+(i%2),(2+(i/2)));
		if (channels[i].chan_freq != 0)
			textcolor(COLOR_GREEN);
		else
			textcolor(COLOR_GRAY1);
		if (radio_freq == channels[i].chan_freq)
			revers(1);
		else
			revers(0);
		printf("%d", i);
		if (i == 0)
			break;
		i++;
		if ( i == MAX_CHAN_AMOUNT)
			i = 0;
	}

}
/* --------------------------------------------------------------- */
static void put_petscii(u8 a_char)
{
    a_char_to_display = a_char;
    __asm__ ("lda %v", a_char_to_display);
    __asm__ ("jsr $FFD2");
}

/* --------------------------------------------------------------- */
static u8 init_fmradio()
{
    u8 hi, lo;

#ifdef VICE
    radio_state = RADIO_ON;
    return OK;
#endif

	//Check chip id
    if (i2c_read_register(RADIOADDR, RDA5807_CHIPID_REG, &hi, &lo) != OK)
    {
    	printf("Failed to read Chip id\n");
    	return NOT_OK;
    }
    if (hi != RDA5807_CHIP_ID)
    {
    	printf("Unknown Chip id 0x%x \n", hi);
    	return NOT_OK;
    }

#if 0
    //Resetting doesn't seem to be having on effect on anything really - skipped for now...
    //Reset the chip
 	if (i2c_update_register(RADIOADDR, RDA5807_CTRL_REG, 0x00, RDA5807_MASK_CTRL_LOBYTE_SOFTRESET,
 			0x00, RDA5807_MASK_CTRL_LOBYTE_SOFTRESET) != OK)
     {
     	printf("Failed to reset Chip\n");
     	return NOT_OK;
     }
 	time_delay();
#endif

    if (set_radio_state(RADIO_ON) != OK)
    {
    	printf("Failed to enable radio\n");
    	return NOT_OK;
    }
    if (set_deemphasis(deemphasis) != OK)
    {
    	printf("Failed to set de-emphasis\n");
    	return NOT_OK;
    }
    if (set_frequency(&radio_freq) != OK)
    {
    	printf("Failed to set frequency!\n");
    	return NOT_OK;
    }

    if (set_mute(OFF) != OK)
    {
    	printf("Failed to unmute!\n");
    	return NOT_OK;
    }
    if (set_volume(volume) != OK)
    {
    	printf("Failed to set volume!\n");
    	return NOT_OK;
    }
    radio_state = RADIO_ON;
    return (OK);
}

/* --------------------------------------------------------------- */
static u8 set_radio_state(u8 state)
{
	u8 ret, data = 0;

#ifdef VICE
    return OK;
#endif

	if (state == RADIO_ON)
	{
		//data = RDA5807_MASK_CTRL_LOBYTE_ENABLE | RDA5807_MASK_CTRL_LOBYTE_SOFTRESET;
		data = RDA5807_MASK_CTRL_LOBYTE_ENABLE;
#if 0
		//Reset the chip
		if (i2c_update_register(RADIOADDR, RDA5807_CTRL_REG, 0x00, RDA5807_MASK_CTRL_LOBYTE_SOFTRESET,
				0x00, RDA5807_MASK_CTRL_LOBYTE_SOFTRESET) != OK)
	    {
	    	printf("Failed to reset Chip\n");
	    	return NOT_OK;
	    }
		time_delay();
#endif
	}
	else
		update_vert_bar(0, RSSI_BAR_X, RSSI_BAR_Y); // Clear RSSI

	ret = i2c_update_register(RADIOADDR, RDA5807_CTRL_REG, RDA5807_MASK_CTRL_HIBYTE_DHIZ, RDA5807_MASK_CTRL_LOBYTE_ENABLE,
			RDA5807_MASK_CTRL_HIBYTE_DHIZ, data);
	if (ret != OK)
		return ret;

	if (state == RADIO_ON)
	return i2c_update_register(RADIOADDR, RDA5807_CHANNEL_REG, 0x00, RDA5807_MASK_CHAN_TUNE,
			0x00, RDA5807_MASK_CHAN_TUNE);

	return ret;
}

/* --------------------------------------------------------------- */
static u8 set_mute(u8 state)
{
	u8 data = RDA5807_MASK_CTRL_HIBYTE_DMUTE;

#ifdef VICE
    return OK;
#endif

	if (state == MUTE_ON)
		data = 0;

	return i2c_update_register(RADIOADDR, RDA5807_CTRL_REG, RDA5807_MASK_CTRL_HIBYTE_DMUTE, 0x00,
						data, 0x00);
}

/* --------------------------------------------------------------- */
static u8 set_frequency(u16 *freq_100khz)
{
	u16 mask = 0;
	u16 val = 0;

	if (*freq_100khz < RDA5807_FREQ_MIN_100KHZ)
		*freq_100khz = RDA5807_FREQ_MAX_100KHZ;
	if (*freq_100khz > RDA5807_FREQ_MAX_100KHZ)
	    *freq_100khz = RDA5807_FREQ_MIN_100KHZ;

#ifdef VICE
    return OK;
#endif

	/* select 50 kHz channel spacing */
	mask |= RDA5807_MASK_CHAN_SPACE;
	val  |= 2 << RDA5807_SHIFT_CHAN_SPACE;

	/* select frequency */
	mask |= RDA5807_MASK_CHAN_WRCHAN;
	val  |= ((((*freq_100khz - RDA5807_FREQ_MIN_100KHZ) * 100) + 25) / 50)
	     << RDA5807_SHIFT_CHAN_WRCHAN;
	/* start tune operation */
	mask |= RDA5807_MASK_CHAN_TUNE;
	val  |= RDA5807_MASK_CHAN_TUNE;

	return i2c_update_register(RADIOADDR, RDA5807_CHANNEL_REG, (mask>>8), (mask & 0xFF),
						(val>>8), (val & 0xFF));
}

/* --------------------------------------------------------------- */
static u16 get_frequency()
{
	u16 val;
	u8 hi, lo;

#ifdef VICE
    return 1048;
#endif

	i2c_read_register(RADIOADDR, RDA5807_CHANNEL_REG, &hi, &lo);
	val = hi;
	val = val << 8;
	val = val | lo;

	val = val >>  RDA5807_SHIFT_CHAN_WRCHAN;
	return (u16)((((50 * val) - 25) / 100) + 1 + RDA5807_FREQ_MIN_100KHZ);

}

/* --------------------------------------------------------------- */
static u8 set_volume(u8 vol)
{
	if (vol == 255)
		vol = RDA5807_VOLUME_MIN;
	else if (vol > RDA5807_VOLUME_MAX)
		vol = RDA5807_VOLUME_MAX;
	volume = vol;

#ifdef VICE
    return OK;
#endif

	return i2c_update_register(RADIOADDR, RDA5807_INTM_THRESH_VOL_REG,
								00, RDA5807_MASK_LOBYTE_VOLUME_DAC,
								0x00, vol);
}

/* --------------------------------------------------------------- */
static u8 toggle_bass_boost()
{
	u8 data = RDA5807_MASK_CTRL_HIBYTE_BASS;

	bass_boost ^= 1;
	if (bass_boost == OFF)
		data = 0;
#ifdef VICE
    return OK;
#endif

	return i2c_update_register(RADIOADDR, RDA5807_CTRL_REG,
				RDA5807_MASK_CTRL_HIBYTE_BASS, 0x00,
				data, 0x00);
}

/* --------------------------------------------------------------- */
static u8 set_stereo_state(u8 set_stereo)
{
	u8 data = SET_STEREO;

	if (set_stereo == SET_MONO)
		data = RDA5807_MASK_CTRL_HIBYTE_MONO;

#ifdef VICE
    return OK;
#endif

	return i2c_update_register(RADIOADDR, RDA5807_CTRL_REG,
			RDA5807_MASK_CTRL_HIBYTE_MONO, 0x00,
			data, 0x00);
}

/* --------------------------------------------------------------- */
static u8 set_deemphasis(u8 deemphasis)
{

#ifdef VICE
    return OK;
#endif

	return i2c_update_register(RADIOADDR, RDA5807_IOCFG_REG,
			RDA5807_MASK_HIBYTE_DEEMPHASIS, 0x00,
			deemphasis, 0x00);
}

/* --------------------------------------------------------------- */
#ifdef ADDFEATURES

static u8 set_lna_port(u8 lna_port)
{

#ifdef VICE
    return OK;
#endif

	return i2c_update_register(RADIOADDR, RDA5807_INTM_THRESH_VOL_REG,
			0x00, RDA5807_MASK_LOBYTE_LNA_PORT,
			0x00, lna_port);
}

/* --------------------------------------------------------------- */
static u8 set_lna_current(u8 lna_current)
{

#ifdef VICE
    return OK;
#endif

	return i2c_update_register(RADIOADDR, RDA5807_INTM_THRESH_VOL_REG,
			0x00, RDA5807_MASK_LOBYTE_LNA_CURRENT,
			0x00, lna_current);
}
#endif

/* --------------------------------------------------------------- */
static u8 get_rssi_and_stat()
{
	u8 hi, lo;

#ifdef VICE
	rssi = 63;
	fm_is_channel = 0x01;
	fm_seek_ready = 0x80;
    return OK;
#endif

	if (i2c_read_register(RADIOADDR, RDA5807_SIGNAL_REG, &hi, &lo) == OK)
		{
		rssi = hi;
		rssi = rssi >> 1;
		fm_is_channel = hi & 0x01;
		fm_seek_ready = lo & 0x80;
		return OK;
		}
	else
		return NOT_OK;
}

/* --------------------------------------------------------------- */
static u8 get_stc_sf_and_st()
{
	u8 hi, lo;

#ifdef VICE
	seek_or_tune_complete = 0x01;
	stereo_indicator = STEREO;
    return OK;
#endif

	if (i2c_read_register(RADIOADDR, RDA5807_SEEK_RESULT_REG, &hi, &lo) == OK)
		{
		seek_or_tune_complete = (hi & RDA5807_MASK_SEEKRES_HIBYTE_COMPLETE) >> 7;
		if (hi & RDA5807_MASK_SEEKRES_HIBYTE_FAIL)
			seek_fail = NOT_OK;
		else
			seek_fail = OK;
		if (hi & RDA5807_MASK_SEEKRES_HIBYTE_STEREO)
			stereo_indicator = STEREO;
		else
			stereo_indicator = MONO;
		return OK;
		}
	else
		return NOT_OK;
}

/* --------------------------------------------------------------- */
static u8 start_scan(u8 direction)
{
#ifdef VICE
    return OK;
#endif

	return i2c_update_register(RADIOADDR, RDA5807_CTRL_REG,
			RDA5807_MASK_CTRL_HIBYTE_SEEKUP | RDA5807_MASK_CTRL_HIBYTE_SEEK, 0x00,
				(direction << 1) | 1, 0x00);
}

/* --------------------------------------------------------------- */

static void start_timer()
{
    /* C64 1MHz gives about 1us for a timer click. If we use both LOW and HIGH bytes
    of TIMER B - then we've got a 16bit timer --> range from 1us to 65ms.
    We can choose to use only the HIGH byte and set the LOW byte 0xFF which
    gives us the same range with a resolution of 255us */

    /* Stop the timer ... in case it would still be running for some reason. */
    POKE (TIMERBCONTROL, 0);

    POKE (TIMERBLOW, 0xff); /* Low byte must be set first !!! */
    POKE (TIMERBHIGH, 0xff);
    POKE (TIMERBCONTROL, 9); /* Start the timer */
}

/* --------------------------------------------------------------- */
static u8 is_timer_complete(u8 ticks)
{
	if ((PEEK(TIMERBCONTROL)&TIMERSTARTBIT_MASK) == 0)
		{
		start_timer();
		timer_ticks++;
		if (timer_ticks >= ticks)
			{
			timer_ticks = 0;
			return (TIMER_EXPIRED);
			}
		}
	return (TIMER_RUNNING);

}

/* --------------------------------------------------------------- */
static void print_registers()
{
	u8 regs[7] = { 0x00, 0x02, 0x03, 0x04, 0x05, 0x0A, 0x0B };
	u8 i = 0, hi, lo;

	gotoxy(0,15);
	textcolor(COLOR_GREEN);
	for (i = 0; i < 7; i++)
	{
		i2c_read_register(RADIOADDR, regs[i], &hi, &lo);
		printf("REG:0x%02x = 0x%02x%02x\n", regs[i], hi, lo);
	}

}

/* --------------------------------------------------------------- */
static u8 save_config_file()
{
    FILE *fp;
    int i=0;
    u8 t_chars[5];

    if ((fp = fopen("fmradio.cfg", "w")) != NULL)
        {
        while (i < MAX_CHAN_AMOUNT)
            {
        	if (channels[i].chan_freq == 0)
        	{
                i++;
        		continue;
        	}

        	if (fputs(CFG_CHA_MAGIC":", fp) ==  EOF)
        		goto func_error;
            ltoa(channels[i].chan_number, t_chars, 10);
            if (fputs(t_chars, fp) == EOF)
            	goto func_error;
            if (fputc(':', fp) == EOF)
            	goto func_error;
            ltoa(channels[i].chan_freq, t_chars, 10);
            if (fputs(t_chars, fp) == EOF)
            	goto func_error;
            if (fputc(':', fp) == EOF)
            	goto func_error;
            if (fputs(channels[i].name, fp) == EOF)
            	goto func_error;
            if (fputc('\n', fp) == EOF)
            	goto func_error;
            i++;
            }
        fclose(fp);
        }
    else
    	goto func_error;

    return OK;

func_error:
	if (fp)
	{
		fclose(fp);
		return NOT_OK;
	}
}

/* --------------------------------------------------------------- */
static u8 read_config_file()
{
    FILE *fp;
    char line[30];
    u8 *token;
    signed char index;

	gotoxy(0,16);

    if ((fp = fopen("fmradio.cfg", "r")) != NULL)
    {
        while (fgets (line, sizeof(line) - 1, fp) != NULL)
        {
        	token = strtok(line, ":");
    		//printf("1:%s", token);
        	if (strcasecmp(token, CFG_CHA_MAGIC) == 0)
        	{
        		token = strtok(NULL, ":");
        		//printf("2:%s", token);
        		if (token == NULL)
        			goto error;
        		index = atoi(token);
        		if ((index > (MAX_CHAN_AMOUNT - 1)) || (index < 0))
					goto error;
        		channels[index].chan_number = index;

        		token = strtok(NULL, ":");
        		//printf("3:%s", token);
        		if (token == NULL)
        			goto error;

        		channels[index].chan_freq = atoi(token);
        		if ((channels[index].chan_freq < RDA5807_FREQ_MIN_100KHZ) ||
        				(channels[index].chan_freq > RDA5807_FREQ_MAX_100KHZ))
        			goto error;

        		token = strtok(NULL, ":");
        		//printf("4:%s", token);
        		if (token == NULL)
        		    goto error;
        		strncpy(channels[index].name, token, sizeof(channels[index].name));
        		//FIXME - this should not be needed
       			channels[index].name[strlen(channels[index].name)-1] = '\0';
        	}
        	//printf("%s", line);
        }
        fclose(fp);
    }
    else
    {
    	//printf("could not open");
    	return NOT_OK;
    }

    return(OK);

error:
    if (fp)
    	fclose(fp);
    return NOT_OK;
}

/* --------------------------------------------------------------- */
static void tune_to_first_stored_slot()
{
	u8 i = 1;

	for (; i < MAX_CHAN_AMOUNT; i++)
	{
		if (channels[i].chan_freq != 0)
		{
			radio_freq = channels[i].chan_freq;
			set_frequency(&radio_freq);
			break;
		}
	}

}
