/*
 * i2c.c - I2C API for C64
 *
 * Copyright (c) 2013 Timo Voutilainen
 *
 * This file is part of FM Radio for Commodore 64.
 *
 * FM Radio for Commodore 64 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FM Radio for Commodore 64 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FM Radio for Commodore 64.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef I2C_HEADER
#define I2C_HEADER
#include <stdio.h>
#include <peekpoke.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <c64.h>
#include <ctype.h>

#define HIGH	1
#define LOW		0
#define OK 		0x01
#define NOT_OK	0x00
#define WRITEBIT 0x00 /* In I2C Address byte */
#define READBIT 0x01  /* In I2C Address byte */
#define PURE_BYTE 0x02 /* Writer function knows that no R/W to be added */

typedef unsigned char u8;
typedef unsigned int u16;

void time_delay();
u8 init_i2c_bus();

void write_start_bit();
u8 write_stop_bit();
void write_master_ack();
void write_master_nack();

u8 i2c_update_register(u8 i2caddr, u8 reg, u8 himask, u8 lomask, u8 hibyte, u8 lobyte);
u8 i2c_read_byte(void);
u8 i2c_write_a_byte(u8 bytetowrite, u8 rw_bit);
u8 i2c_read_register(u8 i2addr, u8 reg, u8 *hibyte, u8 *lobyte);
u8 i2c_write_register(u8 i2caddr, u8 reg, u8 hibyte, u8 lobyte);

#endif
